/*
 * $Id: gnome-synce-manager.cs 110 2007-04-02 11:05:24Z lloydd $
 */
 
using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using Mono.Unix;
using Mono.GetOptions;
using NDesk.DBus;
using Gtk;
using Glade;
using GLib;
using Gnome.Keyring;
using org.wimo.dccm;

namespace Manager
{
   public enum DeviceEventType {
      None,
      Connected,
      Disconnected,
      ConnectionEstablished,
      WrongPassword,
   }
                                                 
   public class MainClass
   {
      public static Dictionary<string,string> device_connection_file;
      public static Dictionary<string,string> device_local_socket;
      public static ManagerOptions Opt;
      public static IDeviceManager Manager;
      static TrayIconManager tray_icon;
      static DevicePasswordInputDlg psw_dlg;
      static bool password_found_in_keyring;

      public static IDevice FindDevice (string path)
      {
         IDevice dev = null;
         
         if (Opt.Debug)
            Console.WriteLine ("Finding device \"{0}\"", path == null ? "null" : path);
         
         try {
            dev = Device.FindInstance (path);
            
            if (Opt.Debug)
               Console.WriteLine ("Device \"{0}\" {1}found", path, (dev == null ? "not " : ""));
         }
         catch (Exception e) {
            Console.Error.WriteLine ("Exception: {0}", e.Message);
         }

         return dev;
      }

      public static void TrayIconVisible (bool visible)
      {
         if (tray_icon == null && !visible)
            return;
         
         if (tray_icon == null)
            tray_icon = new TrayIconManager ();

         if (!visible)
            tray_icon.Hide ();
         else
            tray_icon.Show ();
      }

      public static void Notify (string summary, string body)
      {
         Notifications.Notification n =
            new Notifications.Notification (summary, body, "gnome-pocketpc");
         n.Timeout = 3000;
         n.Urgency = Notifications.Urgency.Normal;
         n.Show ();
      }
      
      public static void PrintDevice (IDevice dev, string path, DeviceEventType type)
      {
         Console.WriteLine ("---[ {0} ]---", type);
         Console.WriteLine ("Path: {0}", dev != null ? dev.Path : path);
         
         if (dev == null) {
            
            Console.WriteLine ("NULL");
            
            return;
         }
         
         Console.WriteLine ("IP: {0}", dev.DeviceIP);
         Console.WriteLine ("RAPI address: {0}:{1}", dev.DeviceRapiIP, dev.DeviceRapiPort);
         Console.WriteLine ("Id: {0}", dev.Id);
         Console.WriteLine ("Current partner id: {0}", dev.CurrentPartnerId);
         Console.WriteLine ("OS Version: {0}.{1}.{2}", dev.OsVersionMajor, dev.OsVersionMinor, dev.Version);
         Console.WriteLine ("CPU type: {0}", dev.CpuType);
         Console.WriteLine ("Guid: \"{0}\"", dev.Guid);
         Console.WriteLine ("Name: \"{0}\"", dev.Name);
         Console.WriteLine ("Platform name: \"{0}\"", dev.PlatformName);
         Console.WriteLine ("Model name: \"{0}\"", dev.ModelName);
      }
      
      public static string SynceDirectory {
         get {
            string dir = System.Environment.GetEnvironmentVariable ("HOME") + "/.synce";
            
            if (!Directory.Exists (dir))
               Directory.CreateDirectory (dir);
            
            return dir;
         }
      }
      
      public static string SynceRAPI2Directory {
         get {
            string dir = SynceDirectory + "/rapi2";
            
            if (!Directory.Exists (dir))
               Directory.CreateDirectory (dir);

            return dir;
         }
      }
      
      public static string SynceConfigFile (string device_name)
      {
         return SynceDirectory + "/" + device_name.ToLower ();
      }
      
      public static string ActiveConnectionFile {
         get {
            return SynceDirectory + "/active_connection";
         }
      }
      
      public static void DevicePasswordRequestedEvent (string device_path)
      {
         IDevice dev = null;
         
         if (Opt.Debug)
            Console.WriteLine ("Password requested");
         
         if ((dev = FindDevice (device_path)) == null) {
            
            Console.Error.WriteLine ("Device {0} not found", device_path);
            
            return;
         }
         
         Hashtable tbl = new Hashtable ();
         tbl ["name"] = dev.Name;
         tbl ["modelname"] = dev.ModelName;
         
         ItemData[] s = Ring.Find (ItemType.GenericSecret, tbl);
         
         password_found_in_keyring = s.Length == 1;
         
         if (s.Length == 1) {
            
            if (Opt.Debug)
               Console.WriteLine ("Password found in GNOME Keyring");
            
            Ring.DeleteItem (s [0].Keyring, s [0].ItemID);
            
            dev.Password = s [0].Secret;
         } else {
         
            psw_dlg = new DevicePasswordInputDlg (dev.Name, dev.ModelName);
            
            if (psw_dlg.Run ()) {
   
               if (Opt.Debug)
                  Console.WriteLine ("Providing password");
               
               try {
                  dev.Password = psw_dlg.Password;
   
                  if (Opt.Debug)
                     Console.WriteLine ("Password provided");
               }
               catch (Exception e) {
                  
                  Console.Error.WriteLine ("Exception: {0}", e.Message);
               }
            } else
               if (Opt.Debug)
                  Console.WriteLine ("Password not provided");
               
            psw_dlg = null;
         }
      }

      public static void DeviceConnectedEvent (string device_path)
      {
         IDevice dev = null;

         if ((dev = FindDevice (device_path)) == null) {
            Console.Error.WriteLine ("Device {0} not found", device_path);
            return;
         }
         
         if (Opt.Debug)
            PrintDevice (dev, device_path, DeviceEventType.Connected);

         Notify (string.Format (Catalog.GetString ("{0} connected"), dev.PlatformName),
            string.Format(Catalog.GetString ("A {0} {1} '{2}' just connected."),
               dev.ModelName, dev.PlatformName, dev.Name));
      }
      
      public static void DeviceConnectionEstablishedEvent (string device_path)
      {
         IDevice dev = null;

         if ((dev = FindDevice (device_path)) == null) {
            Console.Error.WriteLine ("Device {0} not found", device_path);
            return;
         }
         
         if (Opt.Debug)
            PrintDevice (dev, device_path, DeviceEventType.ConnectionEstablished);

         string pid = UnixProcess.GetCurrentProcess ().Id.ToString ();
         string synce_config_file = SynceConfigFile (dev.Name);
         string device_name = dev.Name + " " + dev.ModelName;
         
         if (Opt.Debug)
            Console.WriteLine ("Creating connection file");
         
         using (TextWriter tw = new StreamWriter (synce_config_file)) {
            tw.WriteLine ("[dccm]");
            tw.WriteLine ("pid={0}", pid);
            tw.WriteLine ();
            tw.WriteLine ("[device]");
            tw.WriteLine ("os_version={0}", dev.OsVersionMajor);
            tw.WriteLine ("build_bumber={0}", dev.Version);
            tw.WriteLine ("build_number={0}", dev.Version);
            tw.WriteLine ("processor_type={0}", dev.CpuType);
            tw.WriteLine ("partner_id_1={0}", dev.CurrentPartnerId);
            tw.WriteLine ("partner_id_2={0}", dev.Id);
            tw.WriteLine ("name={0}", dev.Name.ToLower ());
            tw.WriteLine ("real_name={0}", dev.Name);
            tw.WriteLine ("class={0}", dev.PlatformName);
            tw.WriteLine ("hardware={0}", dev.ModelName);
            tw.WriteLine ("ip={0}", dev.DeviceIP);
            tw.WriteLine ("port=0");
            
            if (dev.PasswordProtected && dev.PasswordProvided) {
               tw.WriteLine ("password={0}", dev.Password);
               tw.WriteLine ("key={0}", dev.Key);
            }
            
            tw.WriteLine ();
            tw.WriteLine ("[connection]");
            tw.WriteLine ("transport={0}", dev.OsVersionMajor > 4 ? "rndis" : "ppp");
            
            if (dev.PasswordProtected && dev.PasswordProvided) {
               tw.WriteLine ("password={0}", dev.Password);
               tw.WriteLine ("key={0}", dev.Key);
            }
            
            tw.Close ();
         }

         if (Opt.Debug)
            Console.WriteLine ("Creating active connection file");
         
         if (!File.Exists (ActiveConnectionFile))
            File.Copy (synce_config_file, ActiveConnectionFile);
         else {
            string line = "";
            string[] sp = null;
            
            using (TextReader tr = new StreamReader (ActiveConnectionFile)) {
               while ((line = tr.ReadLine ()).Length > 0)
                  if (line.StartsWith ("pid=")) {
                     if ((sp = line.Split ('=')).Length == 2)
                        if (sp [1].Trim () != pid)
                           File.Copy (synce_config_file, ActiveConnectionFile, true);
                     break;
                  }
               tr.Close ();
            }
         }

         try {
            
            string local_socket_filename = SynceRAPI2Directory + "/" + dev.DeviceIP;
            
            new RapiForwardServer (local_socket_filename,
               dev.DeviceRapiIP, (int) dev.DeviceRapiPort).Start ();
               
            device_local_socket [device_path] = local_socket_filename;
         }
         catch (Exception e) {
            
            Console.Error.WriteLine ("Failed to open RapiForwardServer. {0}", e.Message);
         }
         
         device_connection_file [device_path] = synce_config_file;

         Notify (string.Format (Catalog.GetString ("Connection established")),
            string.Format(Catalog.GetString ("Established connection with {0} {1} '{2}'."),
               dev.ModelName, dev.PlatformName, dev.Name));

         if (dev.PasswordProtected && dev.PasswordProvided) {

            if (Opt.Debug)
               Console.WriteLine ("Saving password to GNOME Keyring");
            
            Hashtable tbl = new Hashtable ();
            tbl ["name"] = dev.Name;
            tbl ["modelname"] = dev.ModelName;
            
            Ring.CreateItem (null, ItemType.GenericSecret, device_name, tbl, dev.Password, true);
         }
         
         TrayIconVisible (device_connection_file.Count > 0);
      }

      public static void DeviceDisconnectedEvent (string device_path)
      {
         string synce_config_file = "";
         string local_socket = "";

         if (Opt.Debug)
            PrintDevice (null, device_path, DeviceEventType.Disconnected);
         
         try {
            
            if (device_connection_file.ContainsKey (device_path)) {
               
               synce_config_file = device_connection_file [device_path];
               
               device_connection_file.Remove (device_path);

               if (File.Exists (synce_config_file))
                  File.Delete (synce_config_file);
            }
            
            if (device_local_socket.ContainsKey (device_path)) {
               
               local_socket = device_local_socket [device_path];
               
               device_local_socket.Remove (device_path);

               if (File.Exists (local_socket))
                  File.Delete (local_socket);
            }
   
            if (File.Exists (ActiveConnectionFile) && device_connection_file.Count < 1)
               File.Delete (ActiveConnectionFile);

            Notify (Catalog.GetString ("PocketPC disconnected"),
               Catalog.GetString ("Just disconnected."));
         }
         catch {
         }
         
         TrayIconVisible (device_connection_file.Count > 0);
      }

      public static void DeviceWrongPasswordEvent (string device_path)
      {
         if (Opt.Debug)
            PrintDevice (null, device_path, DeviceEventType.WrongPassword);

         Notify (Catalog.GetString ("PocketPC wrong password"),
            Catalog.GetString ("Just disconnected."));
         
         TrayIconVisible (device_connection_file.Count > 0);

         if (password_found_in_keyring) {
            
            MessageDialog md = new MessageDialog (null,
                                      DialogFlags.Modal,
                                      MessageType.Error,
                                      ButtonsType.Close,
                                      Catalog.GetString ("In GNOME KeyRing was saved wrong password.\n\nPlease try to connect the device again.\n\nYou will be asked for password."));
     
            md.Run ();
            md.Destroy ();
         }
      }
      
      public static int Main (string[] args)
      {
         Utilities.SetProcessName (Defines.Package);
      
         Catalog.Init (Defines.Package, Defines.LocaleDir);

         if (!org.wimo.Tools.DBusInit ())
            return -1;

         Gtk.Application.Init ();
	 
         Opt = new ManagerOptions ();
         Opt.ProcessArgs (args);

			if (!Ring.Available) {
				Console.WriteLine ("The gnome-keyring-daemon cannot be reached.");
				return -3;
			}
         
         device_connection_file = new Dictionary<string,string> ();
         device_local_socket = new Dictionary<string,string> ();
         
         try {
            Manager = DeviceManager.FindInstance ();
         }
         catch (Exception e) {
            Console.Error.WriteLine ("Exception: {0}", e.Message);
         }
         
         if (Manager == null) {
            Console.Error.WriteLine (
               Catalog.GetString ("Failed to find instance of DeviceManager"));
            
            return -2;
         }
	 
         Manager.DeviceConnected += DeviceConnectedEvent;
         Manager.DeviceDisconnected += DeviceDisconnectedEvent;
         Manager.DeviceConnectionEstablished += DeviceConnectionEstablishedEvent;
         Manager.DevicePasswordRequested += DevicePasswordRequestedEvent;
         Manager.DeviceWrongPassword += DeviceWrongPasswordEvent;
         
         if (Opt.Verbose)
            Console.WriteLine (Catalog.GetString ("Running") + (Opt.Debug ? " (debug)" : ""));

         Gtk.Application.Run ();
	 
         return 0;
      }
   }
}